BUFFER_SIZE = 100
ELEMENT_BUFFER_SIZE = 10

# HISTORY KEYS

INITIAL_CLASS_NAME = "initial_class_name"
OUT_CLASS_NAME = "out_class_name"
INITIAL_TIMESTAMP = "initial_timestamp"
OUT_TIMESTAMP = "out_timestamp"
FINITE_STATE_MACHINE = "finite_state_machine"
TRACK_LENGTH = "track_length"
